/**
 * 
 */
package de.mpc.prolic.misc;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.TreeSet;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.mpc.prolic.exceptions.MissingMemberException;
import de.mpc.prolic.exceptions.ReadIniSectionException;
import de.mpc.prolic.model.Biomolecule;
import de.mpc.prolic.model.BiomoleculeGroup;
import de.mpc.prolic.model.BiomoleculeGroupList;
import de.mpc.prolic.model.BiomoleculeList;
import de.mpc.prolic.model.Peptide;
import de.mpc.prolic.model.Protein;


/**
 * Static class that manages all the ProLiC related input and output
 * @author Michael Kohl
 */
public class IO {

	private final static ResourceBundle bundleMess = ResourceBundle.getBundle("de.mpc.prolic.resources.MessageBundle");		//A localized bundle for user information and for logging
	private static final Logger log4j = LogManager.getLogger(IO.class.getName());
	private static final boolean debugMode = false;
	private final static String spacer = " ";
	
	/**
	 * Reading the settings for ProLiC from a specific section of the configuration file.
	 * In case of doShutdown==true (i.e. if reading the specified section of the .ini file is required for an appropriate execution of ProLiC),
	 * a termination of ProLiC will be initiated.
	 * @param iniFile
	 * @param sectionName
	 * @param success 
	 * @return a map instance with key - value pairs for this section.
	 */
	public static Map<String, String> readSectionFromIni(String iniFile, String sectionName, Boolean success, String mess, boolean doShutdown) 
	{
		Map <String,String> aSection= new HashMap <String,String> ();
		if(debugMode){
			String userdir = System.getProperty("user.dir"); 
			System.out.println(userdir);
		}
		FileInputStream fi;
		InputStreamReader isr;
		try {
			fi =new FileInputStream(iniFile);
			isr = new InputStreamReader(fi);
			IniReader iniR=new IniReader(isr);
			//sections=iniR.getMap();
			aSection=(Map<String, String>) iniR.getMap().get(sectionName);
			if (aSection.isEmpty()){				
				throw new ReadIniSectionException(mess);
			}
		}catch (FileNotFoundException e){	
			if(doShutdown){
				IO.fatalErrorMessAndShutdown(mess);
			}else{
				System.out.println(mess);
				log4j.error(mess);
			}
		}catch (IOException ioE){
			if(doShutdown){
				IO.fatalErrorMessAndShutdown(mess);
			}else{
				System.out.println(mess);
				log4j.error(mess);
			}
		}catch (ReadIniSectionException e) {
			if(doShutdown){
				IO.fatalErrorMessAndShutdown(e.getMessage());
			}else{
				System.out.println(e.getMessage());
				log4j.error(e.getMessage());
			}
		}finally{
			
		}
		return aSection;
	}	

	
	/**
	 * Reads protein data (including the related peptide information) from a file in the ProLiC format and builds the protein instances. 
	 * All the proteins specified in the file are collected within an instance of BiomoleculeList, which is returned by the method.
	 * Protein name and gene name information is optional. Consideration of protein and gene names can be turned off with a setting 
	 * entry in the configuration file.
	 * @param file_
	 * @param delimiter_
	 * @param origin_
	 * @param considerProteinNames
	 * @param considerGeneNames
	 * @param log4j
	 * @param inconsistenciesExist
	 * @return an instance of BiomoleculeList that contains all proteins obtained from a specific input file.
	 */
	public static BiomoleculeList readProLiCformatedProteinFile(File file_, String delimiter_, String origin_, 
			boolean considerProteinNames, boolean considerGeneNames, Logger log4j, Boolean inconsistenciesExist){
		BiomoleculeList aBML = new BiomoleculeList();
		FileInputStream fiStream;
		InputStreamReader isReader;
		BufferedReader bufReader;
		String line;
		Protein aProt;
		Peptide aPept;
		String aPeptSeq;
		String[] lineSplit;
		String biomolIdentifier;
		int currColIndex;
		aBML.setListName(origin_);
		try {
			fiStream = new FileInputStream(file_);
			isReader = new InputStreamReader(fiStream);
			bufReader = new BufferedReader(isReader);
			bufReader.readLine(); //the first line is discarded. It is assumed that this line contains header information
			line = bufReader.readLine(); 
			while(line!=null){
				currColIndex=0;
				aProt = new Protein();
				aProt.setOrigin(origin_);
				lineSplit = line.split(delimiter_);
//The first columns contain protein identifiers, the protein name, the gene name and the protein sequence information
				biomolIdentifier=lineSplit[currColIndex];
				aProt.setIdentifier(biomolIdentifier);
				aProt.setUniqueIdentifier(biomolIdentifier+"_"+origin_);
				currColIndex++;
				if(currColIndex<lineSplit.length)
					aProt.setBiomoleculeType(lineSplit[currColIndex]);
				currColIndex++;
				if (considerProteinNames){
					if(currColIndex<lineSplit.length)
						aProt.setProteinName(lineSplit[currColIndex]);
					currColIndex++;
				}
				if (considerGeneNames){
					if(currColIndex<lineSplit.length)
						aProt.setGeneName(lineSplit[currColIndex]);
					currColIndex++;
				}
				if(currColIndex<lineSplit.length)
					aProt.setSequence(lineSplit[currColIndex]);
				currColIndex++;
//Reading the peptides. The ProLiC format requires that all further columns contain peptide information.				
				for (int i=currColIndex;i<lineSplit.length;i++){
					aPeptSeq=lineSplit[i];
					if (!aPeptSeq.trim().equals("")){
						aPept = new Peptide(lineSplit[i]);
						aProt.addPeptide(aPept);
					}
				}
				aBML.addBiomolecules(aProt);
				line = bufReader.readLine();
			}
			bufReader.close();
		} catch (FileNotFoundException e) {
			log4j.warn(bundleMess.getString("m_fnf_exception_read_input")+". "+bundleMess.getString("m_warning_corrupted_results"));
			System.out.println(bundleMess.getString("m_fnf_exception_read_input")+". "+bundleMess.getString("m_warning_corrupted_results"));
			inconsistenciesExist=true;
		} catch (IOException e) {
			log4j.warn(bundleMess.getString("m_ioe_exception_read_input")+". "+bundleMess.getString("m_warning_corrupted_results"));
			System.out.println(bundleMess.getString("m_ioe_exception_read_input")+". "+bundleMess.getString("m_warning_corrupted_results"));
			inconsistenciesExist=true;
		}
		return aBML;
	}
	

	/**
	 * Writing the results of the comparison to a file. There is only one result file that includes each calculated intersection and each set difference.
	 * @param resultLists
	 * @param filename
	 * @param extension
	 * @param output_dir
	 * @param columnDelim
	 * @param writeProteinNames
	 * @param writeGeneNames
	 * @param writePeptideSeq
	 * @param maxNumbPeptides
	 */
	public static void writeResults(TreeSet<BiomoleculeGroupList> resultLists, String filename, String extension, String output_dir, 
			String columnDelim, boolean writeProteinNames, boolean writeGeneNames, boolean writePeptideSeq, int maxNumbPeptides, 
			Logger log4j, Boolean inconsistenciesExist, String encoding){
		File resFile;
		FileOutputStream oS;
		OutputStreamWriter osw;
		BufferedWriter bw=null;
		String lineSep=System.getProperty("line.separator");
		String outStringData;
		Iterator<BiomoleculeGroupList> itBGL = resultLists.iterator();
		String currListIdentifier;
		ArrayList<BiomoleculeGroup> bgs;
		Iterator<Biomolecule> itB;
		BiomoleculeGroupList aBGL;
		BiomoleculeGroup aBG;
		Biomolecule aB;		
		String BiomoleculeType;
	/*Preparation of the header (TODO for future use - CAVE: further versions of ProLiC that consider different flavors of biomolecules need
	 * a more sophisticated preparation of the header, because there is maybe no protein related information. This is in particular relevant 
	 * in case of biomolecule groups that combine different classes of biomolecules. However, the columns, which should be considered within the 
	 * result file can be already specified to some extent with the settings in the configuration file (genericSettings).) */
		resFile=new File(output_dir+filename+"."+extension);
		System.out.println(resFile.toString());
		String header = "Overlap_Komplement"+columnDelim;
		header = header+"BG GroupId"+columnDelim;
		header=header+ "BG Representative"+columnDelim;
		header=header+"Accession"+columnDelim;
		if (writeProteinNames){
			header=header+"Protein name"+columnDelim;
		}
		if (writeGeneNames){
			header=header+"Gene name"+columnDelim;
		}
		header=header+"Sequence"+columnDelim;
		header=header+"Original List";
		if (writePeptideSeq){
			for (int i=0;i<maxNumbPeptides;i++){
				header=header+columnDelim+"Peptide";
			}
		}
		try{
			oS= new FileOutputStream(resFile);
			if (encoding.equals("UTF-8")){//writes the results in an UTF-8 encoded file in order to support special characters
				osw = new OutputStreamWriter(oS,"UTF-8");
			}
			if (encoding.equals("Latin-1")){
				osw = new OutputStreamWriter(oS,"ISO-8859-1");
			}else{
				osw = new OutputStreamWriter(oS);
			}
			bw = new BufferedWriter(osw);
//Writing the data:
			while(itBGL.hasNext()){//Iterating over intersetions and set differences
				aBGL=itBGL.next();
				currListIdentifier=aBGL.getListIdentifier();
				bw.write(currListIdentifier);
				bw.write(lineSep);
//Writing the header for each result list (increments and set differences)
				bw.write(header);
				bw.write(lineSep);
				bgs = aBGL.getBgList();
				for (int i=0; i<bgs.size();i++){
					aBG=bgs.get(i);
					itB=aBG.getBiomoleculeGroup().iterator();
					while(itB.hasNext()){
						outStringData=currListIdentifier+columnDelim; //First column contains the currennt identifier for the overlaps or the complements
						outStringData=IO.writeBGIdentifier(aBG, outStringData, columnDelim);
						outStringData=IO.writeRepresentative(aBG.getRepresentative(), outStringData, columnDelim); 
						aB=itB.next();
						outStringData=IO.writeBMIdentifier(aB, outStringData, columnDelim);
						BiomoleculeType=aB.getClass().toString();
						try{
							if (writeProteinNames){	
								outStringData=IO.writeProteinName(aB, outStringData, columnDelim, BiomoleculeType);
							}
						}catch(MissingMemberException mme){
							outStringData=outStringData+columnDelim;
						}				
						try{
							if (writeGeneNames){						
								outStringData=IO.writeGeneName(aB, outStringData, columnDelim, BiomoleculeType);
							}
						}catch(MissingMemberException mme){
							outStringData=outStringData+columnDelim;
						}
						outStringData=IO.writeBMSequence(aB, outStringData, columnDelim);
						outStringData=IO.writeBMOrigin(aB, outStringData, columnDelim);
						if (writePeptideSeq){
							outStringData=IO.writePeptideInformations(aB, outStringData, columnDelim, BiomoleculeType);
						}
						bw.write(outStringData);
						bw.write(lineSep);
					}					
				}
				bw.write("");//Adding an empty line between the different lists.
			}
		}catch (FileNotFoundException exfnf){
			log4j.warn(bundleMess.getString("m_fnf_exception_write_results")+" "+filename+extension+".");
			System.out.println(bundleMess.getString("m_fnf_exception_write_results")+" "+filename+extension+".");
			inconsistenciesExist=true;
		}
		catch (IOException exIO){
			log4j.warn(bundleMess.getString("m_ioe_exception_write_results")+" "+filename+extension+". "+bundleMess.getString("m_fileCorrupted_write_results"));
			System.out.println(bundleMess.getString("m_ioe_exception_write_results")+" "+filename+extension+". "+bundleMess.getString("m_fileCorrupted_write_results"));
			inconsistenciesExist=true;
		}
		finally{
			try{
				bw.close();
			}
			catch(IOException exIO){
				log4j.warn(bundleMess.getString("m_fileClosing_write_results")+" "+filename+extension+". "+bundleMess.getString("m_fileCorrupted_write_results"));
				System.out.println(bundleMess.getString("m_fileClosing_write_results")+" "+filename+extension+". "+bundleMess.getString("m_fileCorrupted_write_results"));
				inconsistenciesExist=true;
			}

		}
	}
	
	/**
	 * Auxiliary method used to ensure correct output format of the ProLiC result file even in case of wrong user settings, i.e. enabling writing
	 * of protein name information in case of missing information.
	 * @param aB_
	 * @param outStringData_
	 * @param columnDelim_
	 * @param BiomoleculeType_
	 * @return this auxiliary method is necessary to create a well structured ProLiC result file. 
	 */
	public static String writeProteinName(Biomolecule aB_, String outStringData_, String columnDelim_, String BiomoleculeType_){
		/*The method is used in order to facilitate writing of further lines of the result file even in case of occurring exceptions that
		 * concern only a few Biomolecules */
		if(BiomoleculeType_.equals("class de.mpc.prolic.model.Protein")){
				Protein aP;
				aP=(Protein)aB_; 
				if((aP.getProteinName()!=null)&&(!aP.getProteinName().trim().equals(""))){
					outStringData_=outStringData_+aP.getProteinName()+columnDelim_;	
				}else{//missing information
					Protein prot = (Protein)aB_;
					outStringData_=outStringData_+columnDelim_;	
					throw new MissingMemberException(bundleMess.getString("l_mme_protein_name_info")+spacer+prot.getUniqueIdentifier());
				}
		}else{//wrong biomolecule type
			throw new MissingMemberException(bundleMess.getString("l_mme_biomolecule_not_protein"));
		}
		return outStringData_;
	}
	
	/**
	 * Auxiliary method used to ensure correct output format of the ProLiC result file even in case of wrong user settings, i.e. enabling writing
	 * of gene name information in case of missing availability of such information.
	 * @param aB_
	 * @param outStringData_
	 * @param columnDelim_
	 * @param BiomoleculeType_
	 * @return this auxiliary method is necessary to create a well structured ProLiC result file.
	 */
	public static String writeGeneName(Biomolecule aB_, String outStringData_, String columnDelim_, String BiomoleculeType_) throws MissingMemberException{
		/* The method is used in order to facilitate writing of further lines of the result file even in case of occurring exceptions that
		 * concern only a few Biomolecules.*/
		String resString=outStringData_;
		if(BiomoleculeType_.equals("class de.mpc.prolic.model.Protein")){
				Protein aP;
				aP=(Protein)aB_; 
				if((aP.getGeneName()!=null)&&(!aP.getGeneName().trim().equals(""))){
					resString=outStringData_+aP.getGeneName()+columnDelim_;	
				}else{//missing information
					Protein prot = (Protein)aB_;
					resString=outStringData_+columnDelim_;	
					throw new MissingMemberException(bundleMess.getString("l_mme_gene_name_info")+spacer+prot.getUniqueIdentifier());
				}
		}else{//wrong biomolecule type					
			throw new MissingMemberException(bundleMess.getString("l_mme_biomolecule_not_protein"));
		}
		return resString;			
	}
	
	/**
	 * Auxiliary method used to write information concerning the identifier of the biomolecule group
	 * @param aBG_
	 * @param outStringData_
	 * @param columnDelim_
	 */
	public static String writeBGIdentifier(BiomoleculeGroup aBG_, String outStringData_, String columnDelim_){
		String resString=outStringData_;
		String bgId = String.valueOf(aBG_.getId());
		if (bgId==null){
			resString=resString+""+columnDelim_;
		}else{
			resString=resString+bgId+columnDelim_;
		}
		return resString;
	}
	
	/**
	 * Auxiliary method used to write group representative information
	 * @param bgRepr
	 * @param outStringData_
	 * @param columnDelim_
	 * @return this auxiliary method is necessary to create a well structured ProLiC result file.
	 */
	public static String writeRepresentative(Biomolecule bgRepr, String outStringData_, String columnDelim_){
		String resString=outStringData_;
		String reprId=bgRepr.getIdentifier();
		if(reprId==null){
			resString=resString+""+columnDelim_;
		}else{
			resString=resString+reprId+columnDelim_;
		}
		return resString;
	}
	
	/**
	 * Auxiliary method used to write the identifier of the biomolecule
	 * @param aBM
	 * @param outStringData_
	 * @param columnDelim_
	 * @return this auxiliary method is necessary to create a well structured ProLiC result file.
	 */
	public static String writeBMIdentifier(Biomolecule aBM, String outStringData_, String columnDelim_){
		String resString=outStringData_;
		String bmId=aBM.getIdentifier();
		if(bmId==null){
			resString=resString+""+columnDelim_;
		}else{
			resString=resString+bmId+columnDelim_;
		}
		return resString;
	}

	/**
	 * Auxiliary method used to write the sequence of the biomolecule (either amino acid or nucleotide sequences)
	 * @param aBM
	 * @param outStringData_
	 * @param columnDelim_
	 * @return this auxiliary method is necessary to create a well structured ProLiC result file.
	 */
	public static String writeBMSequence(Biomolecule aBM, String outStringData_, String columnDelim_){
		String resString=outStringData_;
		String bmSeq=aBM.getSequence();
		if(bmSeq==null){
			resString=resString+""+columnDelim_;
		}else{
			resString=resString+bmSeq+columnDelim_;
		}
		return resString;
	}
	
	/**
	 * Auxiliary method used to write the origin of the biomolecule (i.e. the list where the biomolecule originate from)
	 * @param aBM
	 * @param outStringData_
	 * @param columnDelim_
	 * @return this auxiliary method is necessary to create a well structured ProLiC result file.
	 */
	public static String writeBMOrigin(Biomolecule aBM, String outStringData_, String columnDelim_){
		String resString=outStringData_;
		String bmOrig=aBM.getOrigin();
		if(bmOrig==null){
			resString=resString+""+columnDelim_;
		}else{
			resString=resString+bmOrig+columnDelim_;
		}
		return resString;
	}
	/**
	 * Auxiliary method used to ensure correct output format of the ProLiC result file even in case of wrong user settings, i.e. enabling writing
	 * of peptide information in case of missing availability of such information.
	 * @param aB_
	 * @param outStringData_
	 * @param columnDelim_
	 * @param BiomoleculeType_
	 * @return this auxiliary method is necessary to create a well structured ProLiC result file.
	 * @throws MissingMemberException 
	 */
	public static String writePeptideInformations(Biomolecule aB_, String outStringData_, String columnDelim_, String BiomoleculeType_) throws MissingMemberException {
		/* The method is used in order to facilitate writing of further lines of the result file even in case of occurring exceptions that
		 * concern only a few Biomolecules.*/
		Peptide aPept;
		ArrayList <Peptide> pepts;
		if(BiomoleculeType_.equals("class de.mpc.prolic.model.Protein")){
			Protein aP;
			aP=(Protein)aB_; 
			if((aP.getPeptides()!=null)&&(aP.getPeptides().size()>0)){
				pepts=aP.getPeptides();
				for (int i=0;i<pepts.size();i++){
					aPept=pepts.get(i);
					if((aPept.getSequence()!=null)&&(!aPept.getSequence().trim().equals(""))){
						if(i==pepts.size()-1){//the last column delimiter is not written
							outStringData_=outStringData_+aPept.getSequence();	
						}else{
							outStringData_=outStringData_+aPept.getSequence()+columnDelim_;	
						}							
					}
				}
			}else{//missing information
				//throw new MissingMemberException();
				Protein prot=(Protein)aB_;
				throw new MissingMemberException(bundleMess.getString("l_mme_peptide_info")+spacer+prot.getProteinName());
			}
		}else{//wrong biomolecule type
			throw new MissingMemberException(bundleMess.getString("l_mme_biomolecule_not_protein"));
		}
		return outStringData_;
	}
	
	
	/**
	 * fatalErrorMess gives information for the user on the console and performs logging along with the shutdown of ProLiC.
	 * This shutdown is carried out in combination with input / output errors that prevent a meaningful execution of ProLiC.
	 */
	public static void fatalErrorMessAndShutdown(String message){
		System.out.println(message);
		System.out.println("exit(1)");
		log4j.fatal(message);
		log4j.fatal("exit(1)");
		System.exit(1);
	}	
}
