/**
 * 
 */
package de.mpc.prolic.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.Set;

/**
 * @author Michael Kohl
 *  This class stores all Biomolecule instances that are related in terms of identity. This identity characteristic is assigned according 
 *  some kind of identity test (either accession based, sequence based, peptide based), which was performed beforehand. The type of 
 *  comparison is stored also.
 */
public class BiomoleculeGroup implements Comparable<Object> {
	/*******************************************************************************
	 * Members
	 ******************************************************************************/	
	private String comparisonType; 						//The comparison method, which computed the identity relationship
	private ArrayList <Biomolecule> biomoleculeGroup;	//List of biomolecules that passed the similarity criteria.
	private Integer id;									//Unique numeric identifier (with respect to the current ProLiC comparison) for this biomolecule group.
	private Biomolecule representative;					//The biomolecule that is choosen as representative for this biomolecule group

	/*******************************************************************************
	 * Constructors
	 ******************************************************************************/
	/**
	 * Empty constructor
	 */
	public BiomoleculeGroup() {
		init();
	}
	
	/**
	 * Specification of the comparison method applied is given with the creation of the  BiomoleculeGroup instance
	 */
	public BiomoleculeGroup(String comparisionType_) {
		this.comparisonType=comparisionType_;
		init();
	}
	
	/**
	 * Basic initialization of a BiomoleculeGroup
	 */
	private void init(){
		biomoleculeGroup = new ArrayList <Biomolecule>();
	}
	
	/**
	 * The method adds a Biomolecule to this biomolecule group. Additionally, this method performs sorting of the biomolecules while adding biomolecule
	 * group members. 
	 * @param aBiomolecule_
	 */
	public void addBiomolecule(Biomolecule aBiomolecule_){
		biomoleculeGroup.add(aBiomolecule_);
		Collections.sort(biomoleculeGroup);		
	}
	

	/**
	 * Adding all at once the biomolecules that belong to this group. Afterwards, sorting of the added biomolecules is carried out. 
	 * @param bmSet
	 */
	public void addBiomolecules(Set <Biomolecule> bmSet){
		Iterator<Biomolecule> it = 	bmSet.iterator();
		while (it.hasNext()){
			this.biomoleculeGroup.add(it.next());
		}
		Collections.sort(biomoleculeGroup);
	}

	@Override
	/**
	 * Supports the comparison of two instances of BiomoleculeGroup with respect to their unique numeric identifiers.
	 */
	public int compareTo(Object o) {
		BiomoleculeGroup bg = (BiomoleculeGroup) o;
		//return this.getId().compareTo(bg.getId());
		return this.getRepresentative().getUniqueIdentifier().compareTo(bg.getRepresentative().getUniqueIdentifier());
	}
	
	
	/*******************************************************************************
	 * Setter and Getter
	 ******************************************************************************/
	public String getComparisonType(){
		return comparisonType;
	}
	
	public void setComparisonType(String comparisonType_){
		comparisonType = comparisonType_;
	}
	
	public ArrayList<Biomolecule> getBiomoleculeGroup(){
		return biomoleculeGroup;
	}
	
	public void setBiomoleculeGroup(ArrayList<Biomolecule> biomoleculeGroup_){
		biomoleculeGroup = biomoleculeGroup_;
	}
	
	public Biomolecule getRepresentative(){
		return representative;
	}
	
	public void setRepresentative(Biomolecule representative_){
		representative = representative_;
	}

	public void setId(Integer id_){
		id=id_;
	}
	
	public Integer getId(){
		return id;
	}
}
